
from DecisionTree import DecisionTree
import copy

class ID3:

    def __init__(self):
        self.tree_type = DecisionTree
    
    def compute_entropy(self,feature,distinct_values):
        """
        Computes the entropy for a feature in a sample.
        """
        raise NameError("Method compute_entropy must be implemented")
        
    def choose_best_feature(self,training_sample, remaining_features):
        """
        Chooses the best feature for all the training_examples
        entropy's value are float and are greater than 0.
        The more entropy's getting closer to 0, the best it is.

        This method must be overwritten if you want to customize
        the algorithm.
        """

        raise NameError("Method choose_best_feature must be implemented")
        
    def stop_condition(self,training_sample, feature):
        """
        Stop condition for the recursion break.

        This method might be overwritten if you want to customize
        the algorithm.
        """
        distinct_values = self.compute_distinct_values(training_sample,feature)
        return len(distinct_values) <= 1
        
    def compute_distinct_values(self,training_sample,feature):
        """
        Computes a dictionnary where KEYS are the different feature's
        values we have in the training_sample and VALUES are the
        number of occurences of the feature's value in the sample.
        """
        distinct_values = {}

        for training_example in training_sample:
            try:
                example_value = training_example.features[feature]
                if distinct_values.has_key(example_value):
                    distinct_values[example_value] = distinct_values[example_value] + 1
                else:
                    distinct_values[example_value] = 1
            except KeyError:
                pass
        return distinct_values
    

    def get_subsample(self,training_sample,feature,val):
        """
        returns the sub-sample of the TRAINING-EXAMPLE which contains
        all cases such that their FEATURE's value is VAL.
        """
        sub_sample = []
        for training_example in training_sample:
            try:
                if training_example.features[feature] == val:
                    sub_sample.append(training_example)
            except KeyError:
                pass
        return sub_sample


    def make_decision_tree(self,training_sample,feature,remaining_features):
        """
        This method implements the ID3 algorithm.
        In most cases, you shouldn't have to overwrite this method.
        """

        # Test if the stop_condition is verified
        if self.stop_condition(training_sample,feature):
            return self.tree_type((feature,training_sample[0].features[feature]))

        # If the stop condition isn't verified but if there's no remaining
        # features, then let's say we can't decide of the feature's value.

        # XXX: Maybe, it will be necessary to go through the remaining
        # attributes to decide.
        if len(remaining_features) == 0:
            return self.tree_type((feature,"Undecidable"))
        best_feature = self.choose_best_feature(training_sample,remaining_features)
        test_node = self.tree_type(best_feature)
        distinct_values = self.compute_distinct_values(training_sample,best_feature)

        # NECESSARY ???
        new_features = copy.copy(remaining_features)
        # Removes the new "best_feature" from the remaining list.
        new_features.pop(new_features.index(best_feature))
        
        for val in distinct_values.keys():
            sub_sample = self.get_subsample(training_sample,best_feature,val)
            new_branch = self.make_decision_tree(sub_sample,feature,new_features)
            test_node.add_branch(new_branch,val)
        return test_node
        
