# Copyright (c) 2000 Logilab S.A. (France). All Rights Reserved. Tous Droits Rservs.
# http://www.logilab.com. This software was not released to the public, you should not
# have it in your possession, please mail to contact@logilab.com.

##
## Definition of a Kab proxy for Narval
##
from narval.lib import *
import os, sys
import string
from stat import *



# Equivalence list for Kab

L = ['URL', 'additionalName','address','birthday','comment',
     'country','deliveryLabel','email','email2','email3','emails',
     'fax','firstname','fn','idAl','keywords','modem','name',
     'namePrefix','org','orgSubUnit','orgUnit','role','state',
     'talk','telephone','title','town','zip']
     
nameEquiList=[('title',),('firstname',),('namePrefix','namepref'),('name','surname')]
addEquiList=[('address','street'),('country',),('state',),('town',),('zip',)]
orgEquiList=[('org','org'),('orgSubUnit','subdpt'),
             ('orgUnit','dpt'),('role','job')]
contEquiList=[('telephone','phone'),('email','email'),
		     ('email2','email'),('email3','email'),('emails','email'),
		     ('fax','fax'),('modem','web')]
EquiList = nameEquiList + addEquiList + contEquiList + orgEquiList
Dic_kab_to_al = {'title':'title','firstname':'firstname','namePrefix':'namepref','name':'surname','address':'street','country':'country','state':'state','town':'town','zip':'zip','org':'org','orgSubUnit':'subdpt',
'orgUnit':'dpt','role':'job','telephone':'phone','email':'email',
'email2':'email','email3':'email','emails':'email',
'fax':'fax','modem':'web'}

# FIXME : problems with emails 123...

Dic_al_to_kab = {'title':'title','firstname':'firstname','namepref':'namePrefix','surname':'name','street':'address','country':'country','state':'state','town':'town','zip':'zip','org':'org','subdpt':'orgSubUnit',
'dpt':'orgUnit','job':'role','phone':'telephone','email':'email',
'fax':'fax','web':'modem'}

## utility functions for  kab <-> al conversions

def dico_to_node(node,type,dic,list = EquiList):
    """ Transforms a disctionnary into a node, using the translations
    in EquiList.
    CAUTION : these not cannot be added 'as is' to the addbook xml tree"""

    doc = node.ownerDocument
    elmt = doc.createElementNS('',type)
    node.appendChild(elmt)
    for couple in list:
        if dic.has_key(couple[0]) and dic[couple[0]]:
            child = doc.createElementNS('',couple[-1])
            child.appendChild(doc.createTextNode(dic[couple[0]]))
            elmt.appendChild(child)

def dico_to_person_node(doc,dic):
    """from a dictionary, create a 'person' node that can be added to addbook xml tree"""

    entry = doc.createElementNS('','person')
    if dic.has_key('idKab'):
        idKab = dic['idKab']
        entry.setAttributeNS(AL_NS,'idKab',idKab)
    if dic.has_key('idAl'):
        idAl = dic['idAl']
        entry.setAttributeNS(AL_NS,'idAl',idAl)
    status = dic['status']
    entry.setAttributeNS(AL_NS,'status',status)
    # Le nom
    dico_to_node(entry,'name',dic,nameEquiList)
    # La date d'anniversaire
    birth = doc.createElementNS('','birthday')
    entry.appendChild(birth)
    try:
        birth.appendChild(doc.createTextNode(dic['birthday']))
    except KeyError:
        pass
    # L'addresse
    dico_to_node(entry,'address',dic,addEquiList)
    # les contacts
    dico_to_node(entry,'contact',dic,contEquiList)
    # La vie professionnelle
    dico_to_node(entry,'org',dic,orgEquiList)
    # Les commentaires
    comment = doc.createElementNS('','comment')
    entry.appendChild(comment)
    return entry

def node_to_dico(node):
    """ Transforme un noeud d'entre en un dictionnaire
    pour permettre de dtecter modifs ventuelles
    """
    print ' > node_to_dico'    
    dico = {}
    name=''
    for elmt in select_many(node,'descendant::text()'):
        if Dic_al_to_kab.has_key(string.strip(elmt.parentNode.nodeName)):
                name =  Dic_al_to_kab[elmt.parentNode.nodeName]
        dico[name] = elmt.data
    dico['idAl'] = node.getAttributeNS(AL_NS,'idAl')
    dico['idKab'] = node.getAttributeNS(AL_NS,'idKab')
    if node.getAttributeNS(AL_NS,'kab'):
        dico['status']='old'
    print ' < node_to_dico'    
    return dico

class Kab:

    def __init__(self,file):
        self.addr=file
        self.entryDict={}
        
    

    def get_kab_entry(self):
        """ Fonction qui regarde ce qu'il y a dans Kab et en fait un dictionnaire
        de dictionnaires contenant toutes les entres enregistres
        """
        print ' > Kab.get_kab_entry'    
        dict={}
        # begin =0 tant qu'on n'a pas pass la balise [Entries]
        begin=0
        file_kab = open(self.addr)
        for line in file_kab.readlines():
            line = string.strip(line)
            if  not line or line[0] == '#' :
                continue                    
            elif line[0] == '[':
                if line == '[Entries]':
                    begin = 1
                elif line == '[END Entries]':
                    break
                elif begin and line[1:4]=='END':
                    if dict.has_key('idAl'):
                        # cette entre est aussi dans alAB.
                        # Mais peut tre modifie depuis
                        dict['status']="old"
                    else:
                        # Cette entre devra tre ajoute  Al
                        dict['status']='new'
                    self.entryDict[dict['idKab']]=dict
                    dict={}
                elif begin:
                    # change de section !
                    dict['idKab'] = line[1:-1]
                
               
            elif begin:
                line = string.split(line,'=')
                dict[line[0]]= line[1][1:-1]
##                if not line[1]=='""':
##                    # On ne garde que les lments de valeur non nulle
##                    dict[line[0]]= string.split(line[1],'"')[1]
##                else: pass
        
        file_kab.close()
        print ' < Kab.get_kab_entry'    
        return self.entryDict

    def get_entries(self,doc):
    
        print ' > Kab.get_entries'    
        list = []
        Dic = self.get_kab_entry()
        for key in Dic.keys():
            list.append(dico_to_person_node(doc,Dic[key]))
        return list
        

    def reload_entry(self,entry_dico):
        """removes all old entries and write new ones"""

        file_kab = open(self.addr,'r+')
        line = file_kab.readline()
        while not string.strip(line) =='[Entries]':
            line = file_kab.readline()
        file_kab.truncate()
        list_of_keys = entry_dico.keys()
        list_of_keys.sort()
        for id in list_of_keys:
            dico = entry_dico[id]
            file_kab.writelines(["  [%s]\n" % (dico['idKab']),
                                 "    # key-value-pairs\n"])
            entry_keys = dico.keys()
            entry_keys.sort()
            for key in entry_keys:
                file_kab.write('    %s="%s"\n' % (key, dico[key]))
            file_kab.write('  [END %s]\n'% (dico['idKab']))
        file_kab.write('[END Entries]\n')
        file_kab.close()
        

    def add_dico_entry(self,entry_dico):
        """adds a dictionnary of entries to the kab database"""

        file_kab = open(self.addr)
        line = file_kab.readline()
        file_temp = open('addressbook.temp','w')
        self.last_id = 0
        for line in file_kab.readlines():
            if string.strip(line) == '[END Entries]': break
           
            else:
                file_temp.write(line)
                if string.strip(line) == '[Entries]': continue
                elif string.strip(line) == '[Config]': continue
                elif string.strip(line)[0]=='[' and \
                   not string.strip(line)[1:4]=='END':
                    self.last_id =string.strip(line)[1:-1]
        for id in entry_dico.keys():
            dico = entry_dico[id]
            self.last_id = str(int(self.last_id)+1)
            file_temp.writelines(["  [%s]\n" % (self.last_id),
                                 "    # key-value-pairs\n"])
            
            for key in L:
                if dico.has_key(key):
                    file_temp.write('    %s="%s"\n' % (key, dico[key]))
                else:
                    file_temp.write('    %s=""\n'% (key) )
            file_temp.write('  [END %s]\n'% (self.last_id))
        file_temp.write('[END Entries]\n')
        file_temp.close()
        file_kab.close()
        
        # rewrite the file with markup
        file_temp = open('addressbook.temp')
        file_kab = open(self.addr,'w')
        lines = file_temp.readlines()
        file_kab.writelines(lines)
        file_kab.close
        file_temp.close
            

    def add(self,node):
        """ add an entry to Kab
        """

        if node.nodeName=='person':
            dico = node_to_dico(node)
            line = ''
            self.add_dico_entry({dico['idAl']:dico})
            node.setAttributeNS(AL_NS,'idKab',self.last_id)
        else:
            pass

    def add_many(self, list):
        pass
        
    
    def remove(self,node):
        """ remove an entry from kab
        """

        if node.nodeName=='person':
            dico = node_to_dico(node)
            entries = self.get_kab_entry()
            idKab = self.find_entry_id(entries,dico)
            if idKab != None:
                del entries[idKab]
            self.reload_entry(entries)

    def find_entry_id(self,entries,dico):
        for k,v in entries.items():
            if v['name'] == dico['name']:
                match=1
                for field,value in dico.items():
                    if (v.has_key(field) and value == v[field]) or \
                       (value == '' and not v.has_key(field)):
                        pass
                    else:
                        if field[0:2] != 'id':
                            match = 0
                            break
                if match:
                    return k
        return None
                

    
    def modify(self,node_after,node_before):
        self.remove(node_before)
        self.add(node_after)

          
  
    def update(self,tree):
        """ updates kab from Narval addressbook"""

        file_kab = open(self.addr)
        file_temp = open('addressbook.temp','w')
        last_id = 0
        dico_al = {}
        for entry in select_many(tree,'person'):
            dic = node_to_dico(entry)
            dico_al[dic['idAl']]='dic'
        dico_kab = self.get_kab_entry()
        # FIXME: check that all keys are identical in both dictionnaries
        keys_list = dico_al.keys()+dico_kab.keys()
        for id in keys_list:
            if keys_list.count(id)=='2':
                keys_list.remove(id)
            else:
                dico_kab[id] = dico_al[id]
        for dico in dico_kab:
            node = dico_to_person_node(dico,type,dic,list=EquiList)
            self.add(node)

   
        
if __name__ == "__main__" :
    node='''<person idAl="12" idKab="4">
    <name>
      <firstname>fran&#231;oise</firstname>
      <title>Melle</title>
      <surname>PONS</surname>
    </name>
    <birthday/>
    <address>
      <state>none</state>
      <street>none</street>
      <country>none</country>
      <zip>none</zip>
      <town>none</town>
    </address>
    <contact>
      <email>None</email>
      <phone>none</phone>
      <fax>none</fax>
      <web>None</web>
    </contact>
    <type>Person</type>
    <birthday>050478</birthday>
  </person>
'''
    from xml.dom.ext.reader import Sax2
    from xml.dom.ext import StripXml
    person=StripXml(Sax2.FromXml(node)).documentElement
    kab=Kab(".kde/share/apps/kab/addressbook.database")
    kab.remove(person)

        
        
