#!/usr/bin/env python

# Copyright (c) 2001 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


# The learning procedure we decided to use is described in this article :
# http://citeseer.nj.nec.com/mitchell94experience.html

import copy

class LearningProcedure:
    """
    Class which supervise the learning procedure.
    The main steps of the process are :
        - Update the performance statistics for each current rule
        - sort out the Training / Test / Window examples from the current DB
        - Update DB (find new rules, outdate irrelevant ones, ... according
          to the last samples). 
    This class should suits to many learning strategies
    """
    
    def __init__(self,w_ex,features,r_manager,id3_type,base):
        self.window_examples = w_ex
        self.features = features
        self.r_manager = r_manager
        self.id3 = id3_type
        self.filename = base


    def _update_statistics(self):
        """
        This is a first step of the learning process
        """

        # Load old rules
        self.r_manager.load_rules_from_base(self.filename)
        self.old_rules = self.r_manager.rules

        # compute their relevance according to the new examples
        self.r_manager.compute_rules_relevance(window_sample = self.window_examples)
        
    def _sort_examples(self, ratio=2/3.):
        """
        improvement : it would maybe be better to do a shuffle operation
        in the window-sample before computing training-sample and test-sample.
        """
        
        bound = int(ratio*len(self.window_examples))        
        self.training_sample, self.test_sample = self.window_examples[:bound],self.window_examples[bound:]
        

    def _sort_rules_by_relevance(self,rules):
        """
        This method should sort rules and chooses the best ones.
        """

        ################## THIS METHOD IS NOT FINISHED ###################
        training_relevances = map(lambda x: x.training,rules)
        test_relevances = map(lambda x: x.test,rules)

        
    def start(self):
        """
        This is the main method. It calls all the appropriate methods
        in the appropriate order to launch a new learn process.
        """
        
        self._update_statistics()
        self._sort_examples()

        trees = []
        
        for feature in self.features[:]:
            cop = copy.deepcopy(self.features)
            cop.pop(cop.index(feature))
            trees.append(self.id3.make_decision_tree(self.training_sample,feature,cop))
            
        rules = self.old_rules

        for decision_tree in trees:
            rules = rules + decision_tree.compute_rules()
        self.r_manager.set_rules(rules)
        
        self.r_manager.compute_rules_relevance(training_sample=self.training_sample,test_sample=self.test_sample)
        self.r_manager.set_rules(rules)
        self.r_manager.optimize_rules(self.window_examples)
        self.r_manager.remove_redundancy()
        rules = self.r_manager.rules
        rules_to_keep = self._sort_rules_by_relevance(rules)
        self.r_manager.update_base(self.filename)
    
