#!/usr/bin/env python

# Copyright (c) 2001 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


"""
This file contains all the specificities of the meetings
used in the learning processes.
"""

DT_NS = ''
DT_PREFIX = ''
EPSILON = 0.0001

from narval.lib import *


# import TrainingExample # out of Narval
t_example = load_extension_module('pia','TrainingExample')

from Ft.Lib import pDomlette
from Ft.Lib.pDomletteReader import PyExpatReader
from xml.dom.ext import Print, PrettyPrint, StripXml


class Meeting(t_example.TrainingExample):
    """
    Class used to define the specificities of meeting's
    training examples.
    """

    feature_list = ['duration','category','hour','location','time','attendees']
    def __init__(self,features=None,attributes=None):
        self.features = features or {}
        self.remainingAttributes = attributes or {}

    def from_xml(self,vevent):
        """
        Updates the current object with the data contained in
        the vevent (dom version of a VEVENT in vcal format) element.
        """

        hour = None

        # This is the "classic" format.
        try:
            start = select_one(vevent,'DTSTART/text()').data
            hour  = time.strftime("%H%M",time.localtime(_vcal_to_time(utf8_to_latin1(start))))
            # Not sure it's always what we've got to do
            if start[-1] != 'Z':
                start = start + 'Z'
            end = select_one(vevent,'DTEND/text()').data
            # Not sure it's always what we've got to do
            if end[-1] != 'Z':
                end = end + 'Z'
            duraval = int((_vcal_to_time(utf8_to_latin1(end))) - _vcal_to_time(utf8_to_latin1(start))) / 60

        # If we don't have any DTEND/DTSTART elements, let see if we're in the
        # middle of a negociation process.
        except SelectException:
            try:
                start = select_one(vevent,'extension[@x-name="X-NARVAL-DTINTERSTART"]/text()').data
                hour  = time.strftime("%H%M",time.localtime(_vcal_to_time(utf8_to_latin1(start))))
                # Not sure it's always what we've got to do
                if start[-1] != 'Z':
                    start = start + 'Z'
                end = select_one(vevent,'extension[@x-name="X-NARVAL-DTINTEREND"]/text()').data
                # Not sure it's always what we've got to do
                if end[-1] != 'Z':
                    end = end + 'Z'
                
                duraval = select_one(vevent,'extension[@x-name="X-NARVAL-DURATION"]/text()').data
                # duraval = int((_vcal_to_time(utf8_to_latin1(end))) - _vcal_to_time(utf8_to_latin1(start))) / 60
            except SelectException:
                duraval = "0"
                hour = hour or '0000'
            
        self.features["duration"] = duraval
        self.features["hour"] = hour
        
        # Attendees
        attendees = select_many(vevent,'ATTENDEE/text()')
        tple = tuple(map(lambda x: utf8_to_latin1(string.split(x.data,'MAILTO:')[-1]),attendees))
        self.features["attendees"] = tple

        # Location : As we don't directly use it for the moment, we don't
        # really care of it.
        self.features["location"] = "Unknown"

        # category
        try:
            value = unicode(utf8_to_latin1(select_one(vevent,'CATEGORIES/text()').data),'ISO-8859-1')
            self.features["category"] = value
        except SelectException:
            self.features["category"] = "None"
        
        # day
        isodt = _vcal_to_iso8601(start)
        day = weekdays[time.localtime(iso8601.parse(isodt))[6]]
        self.features["day"] = day
        
        for child in vevent.childNodes:
            nodename = child.nodeName
            value = child.firstChild.data
            if nodename not in ['CATEGORIES','ATTENDEE','DTSTART','DTEND']:
                self.remainingAttributes[nodename] = value

        # Make sure we have a valid Meeting format
        self.complete()

    def complete(self):
        """
        Call it to make sure you've got all the features in
        the dictionnary
        """

        for f in ['duration','attendees','hour','location','category','day']:
            if not self.features.has_key(f):
                self.features[f] = ""
        

from Ft.Lib.pDomletteReader import SaxReader
from xml.utils import iso8601


import time, locale

def _vcal_to_iso8601 (vcaldt) :
    return vcaldt[:4] + '-' + vcaldt[4:6] + '-' + vcaldt[6:11] + ':' + \
        vcaldt[11:13] + ':' + vcaldt[13:]

def _vcal_to_time (vcaldt) :
    return iso8601.parse(_vcal_to_iso8601(vcaldt))

import string

weekdays = ['monday','tuesday','wednesday','thursday','friday','saturday','sunday']

def vcal2sample(uri):
    """
    Converts a calendar (vcal format) into a set of
    Meetings.
    """
    import sys

    # REMOVE this when this module is used directly from NARVAL
    # The problem, out of NARVAL, comes from the $NARVAL_HOME
    from imp import load_module, find_module
    sys.path.insert(0,'/home/adim/ypiou')
    mp_file,mp_filename,mp_desc = find_module('extensions')
    mod = load_module('extensions',mp_file,mp_filename,mp_desc)
    mp_file,mp_filename,mp_desc = find_module('xmlrep', mod.__path__)
    mod = load_module('xmlrep',mp_file,mp_filename,mp_desc)
    m_file,m_filename,m_desc = find_module('vcal',mod.__path__)

    final_mod = load_module('vcal',m_file,m_filename,m_desc)

    calendar = final_mod.read(uri)
    
    vevents = select_many(calendar.documentElement,'VEVENT')
    

    sample = []
    for vevent in vevents:
        m = Meeting()
        m.from_xml(vevent)
        sample.append(m)

    return sample


    


dtree_mod = load_extension_module('pia','DecisionTree')

# from DecisionTree import DecisionTree, Rule, RulesManager

class MeetingRule(dtree_mod.Rule):
    """
    This class adds the 'type' functionality.
    """

    def eval(self,meeting):
        """
        If the meeting  matches the rule (self), then evaluates
        if it's positive / negative / ...
        """
        result = 0
        fv_list = meeting.features.items()
        # if all conditions of the rule are verified by the meeting
        if self.match(meeting) >= 0:
            for t in self.consequences.items():
                if (t[0],t[1][0]) in fv_list:
                    if t[1][1] == 'positive':
                        result = result + 1
                    elif t[1][1] == 'negative':
                        result = result - 1
                    elif t[1][1] == 'sure':
                        result = result + 2
                    elif t[1][1] == 'impossible':
                        return impossible
        if result > 0:
            return 'positive'
        elif result < 0:
            return 'negative'
        return ''

    def get_type(self):
        """
        returns the 'type' of the rule according to its performance.
        """
        rate = 0
        if self.window[0]:
            rate = self.window[0]/(self.window[1]*1.)

        if rate > 0.8:
            return 'sure'
        if rate > 0.6:
            return 'positive'
        return ''
    
class MeetingRulesManager(dtree_mod.RulesManager):
    """
    Implements the RulesManager abstract methods (load_rules_from_base
    and update_base)
    """
    
    def xml2rules(self,root_node):
        """
        Converts the xml document to the 'Rule' format.
        """

        # Browses each rules in the xml tree
        for rule in root_node.childNodes:
            attendee_list = []
            
            # STATISTICS
            training = tuple(map(lambda x: int(x),string.split(rule.getAttributeNS(DT_NS,'training-relevance'),'/')))
            test = tuple(map(lambda x: int(x),string.split(rule.getAttributeNS(DT_NS,'test-relevance'),'/')))

            
            condition_nodes = select_many(rule,'if/meetingattr')
            csq_nodes = select_many(rule,'then/meetingattr')
            conditions = {}
            csq = {}

            # CONDITIONS
            for node in condition_nodes:
                # node.firstChild.data <=> select_one(node,'text()').data
                feature = node.getAttributeNS(DT_NS,'name')
                if node.firstChild is None:
                    value = ()
                else:
                    if feature == 'attendees':
                        attendee_list.append(node.firstChild.data)
                        continue
                    else:
                        value = node.firstChild.data
                conditions[feature] = value

            if len(attendee_list):
                # As attendees values are string, then we can sort them
                attendee_list.sort()
                conditions['attendees'] = tuple(attendee_list)
            attendee_list = []
            # CONSEQUENCES
            for node in csq_nodes:
                feature = node.getAttributeNS(DT_NS,'name')
                if node.firstChild is None:
                    value = ()
                else:
                    if feature == 'attendees':
                        attendee_list.append(node.firstChild.data)
                        continue
                        # value = tuple(string.split(node.firstChild.data,', '))
                    else:
                        value = node.firstChild.data
                csq[feature] = value

            if len(attendee_list):
                # As attendees values are string, then we can sort them
                attendee_list.sort()
                csq['attendees'] = tuple(attendee_list)
            # create a new "Rule" object and append it to the list.
            r = MeetingRule(conditions,csq,training,test)
            self.rules.append(r)

    def load_rules_from_base(self,filename):
        """
        Loads the rules from an xml base.
        """
        
        try:
            doc = read_domtree_from_file(filename)
            self.xml2rules(doc.documentElement)
        except IOError:
            self.rules = []
            
    def update_base(self,filename):
        write_domtree_to_file(self.rules2xml(),filename)
        
    
    def rules2xml(self):
        """
        converts the rules dictionnary to an xml document
        """

        # Creates the document Element.
        reader = PyExpatReader()
        doc = reader.fromString("<schedulingrules/>")

        # Creates a "rule" element for each rules we want to store.
        for rule in self.rules:
            rule_node = doc.createElementNS(DT_NS,'prefrule')
            if_node = doc.createElementNS(DT_NS,'if')
            
            # CONDITIONS
            for feature,value in rule.conditions.items():
                if feature == 'attendees':
                    for val in value:
                        condition_node = doc.createElementNS(DT_NS,'meetingattr')
                        condition_node.setAttributeNS(DT_NS,DT_PREFIX+'name',feature)
                        condition_node.appendChild(doc.createTextNode(str(val)))
                        if_node.appendChild(condition_node)
                else:
                    condition_node = doc.createElementNS(DT_NS,'meetingattr')
                    condition_node.setAttributeNS(DT_NS,DT_PREFIX+'name',feature)
                    condition_node.appendChild(doc.createTextNode(str(value)))
                    if_node.appendChild(condition_node)

                    
            values = rule.consequences.values()

            # if one of the csq is "Undecidable", then we don't store the rule.
            if "Undecidable" in values:
                continue
            then_node = doc.createElementNS(DT_NS,'then')

            # CONSEQUENCES
            for feature,value in rule.consequences.items():                
                if feature == 'attendees':
                    for val in value:
                        csq_node = doc.createElementNS(DT_NS,'meetingattr')
                        csq_node.setAttributeNS(DT_NS,DT_PREFIX+'name',feature)
                        csq_node.appendChild(doc.createTextNode(str(val)))
                        csq_node.setAttributeNS(DT_NS,DT_PREFIX+'type',rule.get_type())
                        then_node.appendChild(csq_node)
                else:
                    csq_node = doc.createElementNS(DT_NS,'meetingattr')
                    csq_node.setAttributeNS(DT_NS,DT_PREFIX+'name',feature)
                    csq_node.appendChild(doc.createTextNode(str(value)))
                    csq_node.setAttributeNS(DT_NS,DT_PREFIX+'type',rule.get_type())
                    then_node.appendChild(csq_node)

            rule_node.appendChild(if_node)
            rule_node.appendChild(then_node)

            # STATISTICS
            rule_node.setAttributeNS(DT_NS,DT_PREFIX+'training-relevance',string.join(map(lambda x: str(x),rule.training),'/'))
            rule_node.setAttributeNS(DT_NS,DT_PREFIX+'test-relevance',string.join(map(lambda x: str(x),rule.test),'/'))
            doc.documentElement.appendChild(rule_node)
            
        return doc


import copy


# from ID3 import ID3 # Out of Narval
id3_mod = load_extension_module('pia','ID3')

DEEP_COPY = 1
class MeetingDecisionTree(dtree_mod.DecisionTree):
    """
    meeting-decision-trees use the DecisionTree structure and provide
    a method to compute the associated rules.
    """

    pass



class MeetingID3(id3_mod.ID3):
    """
    Class which implements the abstract methods of the ID3 class
    (choose_best_feature and compute_entropy)
    """
    
    def __init__(self):
        self.tree_type = MeetingDecisionTree
    
    def compute_entropy(self,feature,distinct_values):
        """
        Computes the entropy for a feature in a sample.
        """
        
        from math import log, fabs
        entropy = 0

        nb_examples = reduce(lambda x,y: x+y,distinct_values.values())
        for feature_value, nb_occur in distinct_values.items():
            # force floating division
            ratio = nb_occur/(nb_examples*1.)
            entropy = entropy + (ratio) * log(ratio)
        if fabs(entropy) < EPSILON:
            return 0
        return -entropy    
        
    def choose_best_feature(self,training_sample, remaining_features):
        """
        Chooses the best feature for all the training_examples
        entropy's value are float and are greater than 0.
        The more entropy's getting closer to 0, the best it is.
        """
        
        min = None
        # For all remaining features
        for feature in remaining_features:
            distinct_values = self.compute_distinct_values(training_sample,feature)
            entropy = self.compute_entropy(feature,distinct_values)
            ## print " --> entropy for ",feature, " : ",entropy
            if min is None or entropy < min:
                min = entropy
                best_feature = feature
               
        return best_feature
        


# from LearningProcedure import LearningProcedure

lp_mod = load_extension_module('pia','LearningProcedure')

if __name__ == '__main__':

    if len(sys.argv) == 1:
        training_sample = vcal2sample('/home/adim/ypiou/data/first2.vcs')
        # basename = '/home/adim/apprentissage/modules/test.xml'
        basename = '/home/adim/ypiou/data/schedulingrules.xml'
    elif sys.argv[1] == 'guil':
        training_sample = vcal2sample('/home/adim/apprentissage/modules/guil_cal.vcs')
        basename = '/home/adim/apprentissage/modules/guil_rules.xml'
    elif sys.argv[1] == 'bvf':
        training_sample = vcal2sample('/home/adim/apprentissage/modules/bvf_cal1.vcs')
        basename = '/home/adim/apprentissage/modules/bvf1_rules.xml'
    elif sys.argv[1] == 'bvf2':
        training_sample = vcal2sample('/home/adim/apprentissage/modules/bvf_cal2.vcs')
        basename = '/home/adim/apprentissage/modules/bvf2_rules.xml'
    elif sys.argv[1] == 'alf':
        training_sample = vcal2sample('/home/adim/apprentissage/modules/alf_cal.vcs')
        basename = '/home/adim/apprentissage/modules/alf_rules.xml'
    else:
        training_sample = vcal2sample('/home/adim/ypiou/data/first2.vcs')
        basename = '/home/adim/apprentissage/modules/test.xml'
                
    strategy = MeetingID3()
    r_manager = MeetingRulesManager()
    features = ['duration','category','location','attendees','day','hour']
    
    proc = lp_mod.LearningProcedure(training_sample, features, r_manager, strategy, basename)
    proc.start()
    
