#!/usr/bin/env python

# Copyright (c) 2001 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from Meeting import *

import unittest

class MeetingRuleTestCase(unittest.TestCase):
    """
    This class checks that the MeetingRule's methods give
    the expected result.
    """


    def test_match(self):
        feat1 = {"location":"Logilab",
                 "duration":"120",
                 "category":"RSH",
                 "attendees":("a@b.fr","c@d.fr"),
                 "day":"monday",
                 "hour":"1400"}
        cd1 = {"category":"RSH"}
        csq1 = {"day":"monday"}

        m1 = Meeting(feat1)
        r1 = MeetingRule(cd1,csq1)

        self.assertEqual(r1.match(m1),1)

        csq2 = {"day":"tuesday"}
        r2 = MeetingRule(cd1,csq2)
        self.assertEqual(r2.match(m1),0)

        cd2 = {"category":"Personnel"}

        r2 = MeetingRule(cd2,csq2)
        self.assertEqual(r2.match(m1),None)

        
class MeetingDTTestCase(unittest.TestCase):
    """
    This class checks that the compute_rules' method
    gives the expected result.
    """

    MEETING1 = {
        "category":'Aucune',
        "duration":"15",
        "location":'logilab',
        "attendees":('guil@logilab.fr','syt@logilab.fr','nico@logilab.fr'),
        "day":"Monday",
        "hour":"1500"
        }
    
    MEETING2 = {
        "category":'Aucune',
        "duration":"15",
        "Location":'logilab',
        "attendees":('guil@logilab.fr','syt@logilab.fr','nico@logilab.fr'),
        "day":"Monday",
        "hour":"1500"
        }
    
    MEETING3 = {
        "category":'Aucune',
        "duration":"60",
        "Location":'logilab',
        "attendees":('guil@logilab.fr','syt@logilab.fr','nico@logilab.fr'),
        "day":"Tuesday",
        "hour":"1500"
        }


    DUR_COND2 = {"day":"Tuesday"}
    DUR_COND1 = {"day":"Monday"}

    DUR_CSQ2 = {"duration":"60"}
    DUR_CSQ1 = {"duration":"15"}

    DURATION_RULES = [MeetingRule(DUR_COND2,DUR_CSQ2),MeetingRule(DUR_COND1,DUR_CSQ1)]
    def test_make_dt(self):
        """
        tests creation of decisionTree from meetings
        """

        m1 = Meeting(self.MEETING1,{})
        m2 = Meeting(self.MEETING2,{})
        m3 = Meeting(self.MEETING3,{})
        
        sample = [m1,m2,m3]
        
        strategy = MeetingID3()
        to_test = 'duration'
        remaining_features = ['category','Location','attendees','day','hour']
        tree = strategy.make_decision_tree(sample,to_test,remaining_features)
        rules = tree.compute_rules()
        r_manager = MeetingRulesManager(rules)
        r_manager.optimize_rules(sample)

        if len(r_manager.rules) != len(self.DURATION_RULES):
            raise self.failureException
        for i in range(len(r_manager.rules)):
            if not r_manager.rules[i] == self.DURATION_RULES[i]:
                raise self.failureException


# #############################################################################
    TREE = MeetingDecisionTree("Position")
    leaf1 = MeetingDecisionTree("Hour")

    leaf1_1 = MeetingDecisionTree(("Location","B-390"))
    leaf1_2 = MeetingDecisionTree(("Location","B-391"))
    leaf1.add_branch(leaf1_1,"morning")
    leaf1.add_branch(leaf1_2,"evening")
    
    leaf2 = MeetingDecisionTree(("Location","B-392"))
    leaf3 = MeetingDecisionTree(("Location","B-393"))
    TREE.add_branch(leaf1,"Student")
    TREE.add_branch(leaf2,"Programmer")
    TREE.add_branch(leaf3,"Faculty")


    COND1 = {"Position":'Faculty'}
    COND2 = {"Position":'Student',
             "Hour":'morning'}
    COND3 = {"Position":'Student',
             "Hour":'evening'}
    COND4 = {"Position":'Programmer'}

    CSQ1 = {"Location":"B-393"}
    CSQ2 = {"Location":"B-390"}
    CSQ3 = {"Location":"B-391"}
    CSQ4 = {"Location":"B-392"}
    
    LOCATION_RULES = [MeetingRule(COND1,CSQ1),MeetingRule(COND2,CSQ2),MeetingRule(COND3,CSQ3),MeetingRule(COND4,CSQ4)]

    def test_compute_rules(self):
        """Tests that compute_rule() gives the expected result
        """

        rules = self.TREE.compute_rules()

        if len(rules) != len(self.LOCATION_RULES):
            raise self.failureException
        for i in range(len(rules)):
            if not rules[i] == self.LOCATION_RULES[i]:
                raise self.failureException


    # ##########################################################################
    def test_remove_redudancy(self):
        """Tests that the remove_redundancy methods works fine
        """
        m1 = Meeting(self.MEETING1,{})
        m2 = Meeting(self.MEETING2,{})
        m3 = Meeting(self.MEETING3,{})
        sample = [m1,m2,m3]

        rule1 = MeetingRule(self.COND1,self.CSQ1)
        rule2 = MeetingRule(self.COND2,self.CSQ2)
        rules = [rule1,rule2]

        new_rules = MeetingRule(rule1.conditions,rule1.consequences)
        r_rules = [rule1,rule2,new_rules]
        r_manager = MeetingRulesManager(r_rules)
        r_manager.remove_redundancy()
        for rule in rules:
            if r_manager.rules.count(rule) != 1:
                print "Error on rule : "
                print rule
                raise self.failureException
        


class MeetingRMTestCase(unittest.TestCase):
    """
    Test Class for the MeetingRulesManager
    """

    def test_choose_best_rule(self):
        """Tests if the choose_best_rule method gives the expected result
        """
        r1 = MeetingRule({},{},(5,5),(6,8))
        r2 = MeetingRule({},{},(3,6),(6,9))
        r3 = MeetingRule({},{},(5,5),(7,8))

        rules = [r1,r2,r3]
        r_manager = MeetingRulesManager()
        best_rule = r_manager.choose_best_rule(rules)

        if best_rule.window != (12,13):
            print best_rule.window
            raise self.failureException

        r4 = MeetingRule({},{},(10,10),(20,20))
        r5 = MeetingRule({},{},(0,0),(30,30))
        r6 = MeetingRule({},{},(40,40),(0,0))
        r7 = MeetingRule({},{},(41,42),(51,52))
        rules.append(r4)
        rules.append(r5)
        rules.append(r6)
        rules.append(r7)
        
        best_rule = r_manager.choose_best_rule(rules)

        if best_rule.window != (40,40):
            print best_rule.window
            raise self.failureException
        
    MEETING1 = {
        "category":'Boulot',
        "duration":"105",
        "Location":'logilab',
        "attendees":('guil@logilab.fr','syt@logilab.fr','nico@logilab.fr'),
        "day":"Monday",
        "hour":"1600"
        }
    
    MEETING2 = {
        "category":'Perso',
        "duration":"60",
        "Location":'logilab',
        "attendees":('guil@logilab.fr','syt@logilab.fr','nico@logilab.fr'),
        "day":"Monday",
        "hour":"1600"
        }
    
    MEETING3 = {
        "category":'Aucune',
        "duration":"60",
        "Location":'logilab',
        "attendees":('guil@logilab.fr','syt@logilab.fr','nico@logilab.fr'),
        "day":"Tuesday",
        "hour":"1600"
        }

    MEETING4 = {
        "category":'Aucune',
        "duration":"80",
        "Location":'logilab',
        "attendees":('guil@logilab.fr','syt@logilab.fr','nico@logilab.fr'),
        "day":"Tuesday",
        "hour":"1700"
        }
    
    def test_optimize_rules(self):
        m1 = Meeting(self.MEETING1,{})
        m2 = Meeting(self.MEETING2,{})
        m3 = Meeting(self.MEETING3,{})
        m4 = Meeting(self.MEETING4,{})

        sample = [m1,m2,m3,m4]

        strategy = MeetingID3()
        to_test = 'duration'
        remaining_features = ['category','Location','attendees','day','hour']
        tree = strategy.make_decision_tree(sample,to_test,remaining_features)
        rules = tree.compute_rules()
        r_manager = MeetingRulesManager(rules)
        r_manager.optimize_rules(sample)

        ## TO BE FINISHED ...
        
class MeetingSanityCheck(unittest.TestCase):

    INPUT_FILE = "/home/adim/apprentissage/modules/input.xml"
    OUTPUT_FILE = "/home/adim/apprentissage/modules/output.xml"

    def test_xml_rules(self):
        """Taking rules from base, and update the base without
        any process should not change base"""
        
        r_manager = MeetingRulesManager()
        f = open(self.INPUT_FILE,'r')
        input = f.read()
        f.close()
        r_manager.load_rules_from_base(self.INPUT_FILE)
        old_rules = r_manager.rules
        r_manager.update_base(self.OUTPUT_FILE)

        r_manager.load_rules_from_base(self.OUTPUT_FILE)
        new_rules = r_manager.rules

        self.assertEqual(old_rules,new_rules)

        # We can't compare the two files because we can't foresee the
        # order of the features's dictionnary browsing
##        f = open(self.OUTPUT_FILE,'r')
##        output = f.read()
##        self.assertEqual(input,output)


if __name__ == '__main__':
    testsuite = unittest.TestSuite()
    testsuite.addTest(MeetingDTTestCase("test_compute_rules"))
    testsuite.addTest(MeetingDTTestCase("test_make_dt"))
    testsuite.addTest(MeetingDTTestCase("test_remove_redudancy"))
    testsuite.addTest(MeetingSanityCheck("test_xml_rules"))
    testsuite.addTest(MeetingRMTestCase("test_choose_best_rule"))
    testsuite.addTest(MeetingRMTestCase("test_optimize_rules"))

    testsuite.addTest(MeetingRuleTestCase("test_match"))
    
    test = unittest.TextTestRunner()
    # uncomment next line to write tests results in a file
    #test.__init__(open('unittest.log','w+'))    

    test.run(testsuite)
