# Copyright (c) 2000 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from narval.lib import *
from xml.utils import iso8601
import time
import re
import string


# JUST FOR TESTS !!!!!
# from Meeting import *

meeting_mod = load_extension_module('pia','Meeting')


PIA_NS = ''
PIA_PREFIX = ''

MOD_XML = """<?xml version='1.0' encoding='iso-8859-1'?><module name='pia'>"""


#-------------------- Internal functions and classes -------------------

def _vcal_to_iso8601 (vcaldt) :
    return vcaldt[:4] + '-' + vcaldt[4:6] + '-' + vcaldt[6:11] + ':' + \
        vcaldt[11:13] + ':' + vcaldt[13:]


def _iso8601_to_vcal (isodt) :
    import re
    m = re.match('(\d\d\d\d)-(\d\d)-(\d\dT\d\d):(\d\d)(:(\d\d))?Z(.*)',isodt)
    return m.group(1) + m.group(2) + m.group(3) + m.group(4) + (m.group(6) or '00') + 'Z' + (m.group(7) or '')


def _vcal_to_time (vcaldt) :
    return iso8601.parse(_vcal_to_iso8601(vcaldt))


def _time_to_vcal (time) :
    return _iso8601_to_vcal(iso8601.tostring(time))


def _add_to_date (vcaldt, seconds) :
    '''Takes a VCalendar date (e.g. '20010116T130000Z') and a number of seconds,
    returns the VCalendar date equal to their sum.    
    '''
    from xml.utils import iso8601
    isodt = _vcal_to_iso8601(vcaldt)
    isodt = iso8601.tostring(iso8601.parse(isodt) + seconds)
    return _iso8601_to_vcal(isodt)



weekdays = ['monday','tuesday','wednesday','thursday','friday','saturday','sunday']
DEFAULT_DURATION = 60   # Default duration for events 60 minutes

class VeventCustomizer :
    """Class providing two ways to apply user's scheduling rules on a VEVENT element:
    - complete user's initial request using completion rules from the base;
    - evaluate a proposal using preferences rules from the base.
    """


    def __init__(self, rules):
        self.prefrules = []


        for rule in select_many(rules,'prefrule'):
            conditions = {}
            csqs = {}

            for condition_node in select_many(rule,'if/meetingattr'):
                if condition_node.firstChild:
                    value = str(condition_node.firstChild.data)
                else:
                    value = ""
                conditions[condition_node.getAttributeNS(PIA_NS,'name')] = value
            
            for csq_node in select_many(rule,'then/meetingattr'):

                csq_type = csq_node.getAttributeNS(PIA_NS,'type')
                feature_type = csq_node.getAttributeNS(PIA_NS,'name')
                if csq_node.firstChild:
                    value = csq_node.firstChild.data
                else:
                    value = u""
                csqs[feature_type] = (value,csq_type)

            test_relevance = rule.getAttributeNS(PIA_NS,'test-relevance')
            training_relevance = rule.getAttributeNS(PIA_NS,'training-relevance')
            if test_relevance == '':
                test_relevance = (0,0)
            else:
                test_relevance = tuple(string.split(test_relevance,'/'))

            if training_relevance == '':
                training_relevance = (0,0)
            else:
                training_relevance = tuple(string.split(training_relevance,'/'))
            
            self.prefrules.append(meeting_mod.MeetingRule(conditions,csqs,training_relevance,test_relevance))
            

    def _apply(self, consequences, vevent) :
        '''Modifies the VEVENT as indicated in consequences patterns.
        '''

        try:
            feature_list = map(lambda x: x[0],consequences)
            i = feature_list.index('duration')
            t = consequences.pop(i)
            consequences.insert(0,t)
        except ValueError:
            pass

        doc = vevent.ownerDocument
        for name, value in consequences :
            
            if name == 'attendees':
                value=tuple(string.split(value,','))
                for val in value:
                    if not select_many(vevent, 'ATTENDEE[text()="MAILTO:' + val + '"]'):
                        attendee = doc.createElement('ATTENDEE')
                        vevent.appendChild(attendee)
                        attendee.appendChild(doc.createTextNode('MAILTO:' + val))
                        print val, 'added as an attendee'
                    else:
                        print val, 'was already an attendee'
                    
##            if name == 'attendee' :
##                if not select_many(vevent, 'ATTENDEE[text()="MAILTO:' + value + '"]') :
##                    attendee = doc.createElement('ATTENDEE')
##                    vevent.appendChild(attendee)
##                    text = doc.createTextNode('MAILTO:' + value)
##                    attendee.appendChild(text)
##                    print value, 'added as an ATTENDEE'
##                else :
##                    print value, 'was already an ATTENDEE'
                    
            elif name == 'category' :
                categories = select_many(vevent, 'CATEGORIES/text()')
                if categories :
                    text = categories[0]
                    if not re.search('^|,| '+value+'$|,| ', utf8_to_latin1(text.data)) :
                        text.data = utf8_to_latin1(text.data + ' ' + value)
                        print value, 'added to CATEGORIES'
                    else :
                        print value, 'was already in CATEGORIES'
                else :
                    categories = doc.createElement('CATEGORIES')
                    vevent.appendChild(categories)
                    text = doc.createTextNode(value)
                    categories.appendChild(text)
                    print value, 'added to CATEGORIES'

            elif name == 'day' :
                startnode = select_one(vevent, 'extension[@x-name="X-NARVAL-DTINTERSTART"]/text()')
                now = time.time()
                tstart = _vcal_to_time(utf8_to_latin1(startnode.data))
                # Select the first same day on the week next to the previous location or now.
                add_days = (weekdays.index(value) + 7 - time.localtime(tstart)[6]) % 7
                tstart = tstart + add_days * 86400  # (24h = 86400sec)
                while tstart < now :
                    add_days = add_days + 7
                    tstart = tstart + 604800  # (1week = 604800sec)
                startnode.data = _time_to_vcal(tstart)
                print 'DTINTERSTART overwritten with', startnode.data
                endnode = select_one(vevent, 'extension[@x-name="X-NARVAL-DTINTEREND"]/text()')
                tend = _vcal_to_time(utf8_to_latin1(endnode.data))
                end = _time_to_vcal(tend + add_days * 86400)
                endnode.data = end
                print 'DTINTEREND overwritten with', end
                

            elif name == 'duration' :
                duranode = select_one(vevent,'extension[@x-name="X-NARVAL-DURATION"]/text()')
                duranode.data = value
                print 'DURATION overwritten with', value


            elif name == 'hour' :
                m = re.match('(\d\d\d\d)-(\d\d\d\d)',value)
                if not m :
                    startnode = select_one(vevent, 'extension[@x-name="X-NARVAL-DTINTERSTART"]/text()')
                    start = _vcal_to_time(startnode.data[:9] + value + '00' + startnode.data[15:])
                    while start < time.time() :
                        start = start + 86400  # (24h = 86400sec)
                    startnode.data = _time_to_vcal(start)
                    print 'DTINTERSTART overwritten with', startnode.data
                    endnode = select_one(vevent, 'extension[@x-name="X-NARVAL-DTINTEREND"]/text()')
                    duration = int(select_one(vevent, 'extension[@x-name="X-NARVAL-DURATION"]/text()').data)
                    endnode.data = _time_to_vcal(_vcal_to_time(startnode.data) + duration * 60)
                    print 'DTINTEREND overwritten with', endnode.data
                else :
                    # Set a time interval
                    startnode = select_one(vevent, 'extension[@x-name="X-NARVAL-DTINTERSTART"]/text()')
                    startnode.data = startnode.data[:9] + m.group(1) + '00' + startnode.data[15:]
                    print 'DTINTERSTART overwritten with', startnode.data
                    endnode = select_one(vevent, 'extension[@x-name="X-NARVAL-DTINTEREND"]/text()')
                    endnode.data = startnode.data[:9] + m.group(2) + '00' + startnode.data[15:]
                    print 'DTINTEREND overwritten with', endnode.data

            else :
                raise 'Read an unknown query for VEVENT change named ' + name



    def test_proposal(self,vevent):
        m = meeting_mod.Meeting()
        m.from_xml(vevent)
        result = 0
        
        for t in m.features.items():
            for rule in self.prefrules:
                evaluation = rule.eval(m)
                if evaluation == 'sure':
                    result = result + 2
                elif evaluation == 'positive':
                    result = result + 1
                elif evaluation == 'negative':
                    result = result - 1
                elif evaluation == 'impossible':
                    return 'impossible'
        if result > 0:
            return 'positive'
        elif result == 0:
            return 'neutral'
        return 'negative'

    def complete_proposal(self,vevent):
        
        m = meeting_mod.Meeting()
        m.from_xml(vevent)
        fv_list = []

        # Note that we don't just take the m.features.items() list. We want
        # to keep the order ot the features_list
        features_list = ['category','duration','location','hour','day','attendees']
        fv_list = [(f,m.features[f]) for f in features_list if m.features.has_key(f)]
        
        completed_list = []
        f_completed = []
        for t in fv_list:
            # if already completed, jump to the next round
            if t[0] in f_completed:
                continue

            for rule in self.prefrules:
                rule_conditions = rule.conditions.items()
                if t in rule_conditions:
                    if len(rule_conditions) == 1:
                        for csq in rule.consequences.items():
                            if csq[0] not in f_completed:
                                completed_list.append((csq[0],csq[1][0]))
                                f_completed.append(csq[0])
                    # XXX
                    else:
                        for csq in rule.consequences.items():
                            if csq[0] not in f_completed:
                                completed_list.append((csq[0],csq[1][0]))
                                f_completed.append(csq[0])

        self._apply(completed_list,vevent)
        
                


def _discretize (original_vevent) :
    # Work on a copy
    vevent = original_vevent.cloneNode(DEEP_COPY)
    doc = original_vevent.ownerDocument
    # Extract interval bounds and duration
    starttime = _vcal_to_time(select_one(vevent,'extension[@x-name="X-NARVAL-DTINTERSTART"]/text()').data)
    duration = int(select_one(vevent, 'extension[@x-name="X-NARVAL-DURATION"]/text()').data) * 60
    endlimit = _vcal_to_time(select_one(vevent,'extension[@x-name="X-NARVAL-DTINTEREND"]/text()').data)
    # Discretize
    list = []
    endtime = starttime + duration
    while endtime <= endlimit :
        copy = vevent.cloneNode(DEEP_COPY)
        select_one(copy, 'extension[@x-name="X-NARVAL-DTINTERSTART"]/text()').data = _time_to_vcal(starttime)
        select_one(copy, 'extension[@x-name="X-NARVAL-DTINTEREND"]/text()').data = _time_to_vcal(endtime)
        list.append(copy)
        starttime = starttime + 900  # 15 minutes
        endtime = endtime + 900
    release_domtree(vevent)
    return list

def _release_list (list) :
    for vevent in list :
        release_domtree(vevent)

#------------------ Shared functions -------------------


def first_event_proposal_f(args) :
    '''Function creating the first proposal to be sent by the initiator agent to
    begin negociation.
    '''
    doc = args.ownerDocument
    outputs = doc.createElementNS(PIA_NS,'outputs')
    vevent = select_one(args,'event-request/VEVENT')
    start = select_one(vevent, 'DTSTART')
    
    # Convert DTSTART to DTINTERSTART
    istart = doc.createElement('extension')
    istart.setAttributeNS(PIA_NS,PIA_PREFIX+ 'x-name', 'X-NARVAL-DTINTERSTART')
    startval = select_one(start, 'text()')
    istart.appendChild(startval)
    vevent.appendChild(istart)
    end = select_one(vevent, 'DTEND')
    vevent.removeChild(start)
    release_domtree(start)

    # Convert DTEND to DTINTEREND
    iend = doc.createElement('extension')
    iend.setAttributeNS(PIA_NS,PIA_PREFIX+ 'x-name', 'X-NARVAL-DTINTEREND')
    endval = select_one(end, 'text()')
    iend.appendChild(endval)
    vevent.appendChild(iend)
    vevent.removeChild(end)
    release_domtree(end)

    # Create DURATION
    iend = doc.createElement('extension')
    iend.setAttributeNS(PIA_NS,PIA_PREFIX+ 'x-name', 'X-NARVAL-DURATION')
    duraval = int((_vcal_to_time(utf8_to_latin1(endval.data)) - _vcal_to_time(utf8_to_latin1(startval.data))) / 60)
    iend.appendChild(doc.createTextNode(str(duraval)))
    vevent.appendChild(iend)

    # Add the user as a participant
    usermail = utf8_to_latin1(select_one(args, 'agent-def[@type="self"]/owner/email/text()').data)
    username = select_many(args, 'agent-def[@type="self"]/owner/name/text()')
    if username :
        username = utf8_to_latin1(username[0].data)
    else :
        username = ''
    attendee = doc.createElement('ATTENDEE')
    attendee.appendChild(doc.createTextNode('MAILTO:' + username + ' <' + usermail + '>'))
    attendee.setAttributeNS(PIA_NS,PIA_PREFIX+ 'STATUS', 'TENTATIVE')
    vevent.appendChild(attendee)

    rules_node = select_one(args,'schedulingrules')
    vcust = VeventCustomizer(rules_node)


    # vcust.apply_comprules(vevent)
    # vcust.apply_learnt_rules(vevent,['category'])

    vcust.complete_proposal(vevent)
    
    event_prop = doc.createElementNS(PIA_NS,'event-proposal')
    outputs.appendChild(event_prop)
    event_prop.appendChild(vevent)
    return outputs

MOD_XML = MOD_XML + \
          """<action name='first_event_proposal' func='first_event_proposal_f'>
          <description lang="fr">Crer la proposition d'horaire initiale.</description>
          <description lang="en">Makes the initial proposal for scheduling event.</description>
          <input use='yes'><match>event-request</match></input>
          <input><match>schedulingrules</match></input>
          <input><match>agent-def[@type='self']</match></input>
          <output><match>event-proposal</match></output>
          </action>
          """


def test_proposal_f(args) :
    '''Function crossing the proposal with the existing calendar.
    '''
    doc = args.ownerDocument
    outputs = doc.createElementNS(PIA_NS,'outputs')
    vevent_node = select_one(args, 'message/body/proposal/VEVENT')
    const_node = select_one(args,'constraints')
    pos_node = doc.createElementNS(PIA_NS,'possibilities')
    outputs.appendChild(pos_node)

    set = _discretize(vevent_node)
    for event in set :
        fit = 0
        for const in select_many(const_node, 'VEVENT') :
            if _vcal_to_time(select_one(const, 'DTSTART/text()').data) < \
                _vcal_to_time(select_one(event, 'extension[@x-name="X-NARVAL-DTINTEREND"]/text()').data) \
                and _vcal_to_time(select_one(const, 'DTEND/text()').data) > \
                _vcal_to_time(select_one(event, 'extension[@x-name="X-NARVAL-DTINTERSTART"]/text()').data) :
                break
        else :
            fit = 1
        pos = doc.createElementNS(PIA_NS, 'item')
        pos_node.appendChild(pos)
        text = doc.createTextNode(str(fit))
        pos.appendChild(text)

    _release_list(set)
    return outputs

MOD_XML = MOD_XML + \
          """<action name='test_proposal' func='test_proposal_f'>
          <description lang="fr">Tester la proposition par croisement avec l'agenda de l'utilisateur.</description>
          <description lang="en">Tests the proposal with user's calendar.</description>
          <input use='yes'><match>message[@type='incoming' and header/type/text()='negociation' and header/subject/text()='event scheduling' and body/proposal]</match></input>
          <input use='yes'><match>constraints</match></input>
          <output><match>possibilities</match></output>
          </action>
          """



def judge_proposal_f(args) :
    '''Function using the personnal preferences to judge the proposal.
    '''
    doc = args.ownerDocument
    outputs = doc.createElementNS(PIA_NS,'outputs')
    vevent_node = select_one(args, 'message/body/proposal/VEVENT')
    rules_node = select_one(args,'schedulingrules')
    possib_node = select_one(args,'possibilities')
    eval_node = doc.createElementNS(PIA_NS,'evaluation')
    outputs.appendChild(eval_node)
    startnode = select_one(vevent_node, 'extension[@x-name="X-NARVAL-DTINTERSTART"]/text()')
    endnode = select_one(vevent_node, 'extension[@x-name="X-NARVAL-DTINTEREND"]/text()')
    text = startnode.data + '-' + endnode.data
    eval_node.setAttributeNS(PIA_NS,PIA_PREFIX+ 'interval', text)

    set = _discretize(vevent_node)
    i = 0
    for pos in select_many(possib_node, 'item/text()') :
        if pos.data == '1' :
            vcust = VeventCustomizer(rules_node)
            note = vcust.test_proposal(set[i])
            # vcust.apply_prefrules(set[i])
            
            # note = vcust.apply_prefrules(set[i])
        else :
            note = 'impossible'
        i = i + 1
        eval = doc.createElementNS(PIA_NS, 'item')
        eval_node.appendChild(eval)
        text = doc.createTextNode(str(note))
        eval.appendChild(text)

    _release_list(set)
    return outputs

MOD_XML = MOD_XML + \
          """<action name='judge_proposal' func='judge_proposal_f'>
          <description lang="fr">Juger la proposition avec les prfrences de l'utilisateur.</description>
          <description lang="en">Judges the proposal using user preferences.</description>
          <input use='yes'><match>message[@type='incoming' and header/type/text()='negociation' and header/subject/text()='event scheduling' and body/proposal]</match></input>
          <input use='yes'><match>possibilities</match></input>
          <input><match>schedulingrules</match></input>
          <output><match>evaluation</match></output>
          </action>
          """


def interpret_evaluation_f(args) :
    '''Function to decide if a consensus has been reached or if the
    negociation must continue.
    '''
    doc = args.ownerDocument
    outputs = doc.createElementNS(PIA_NS,'outputs')
    nego_node = select_one(args, 'negociation')
    outputs.appendChild(nego_node)
    vevent_node = select_one(nego_node, 'cycle[last()]/proposal/VEVENT')

    values = []
    for eval_node in select_many(nego_node, 'cycle[last()]/evaluation') :
        values.append(map(lambda x : utf8_to_latin1(x.data), select_many(eval_node, 'item/text()')))

    maxi = None   # To keep the maximum evaluations sum.
    maxon = None   # To keep the item of the place whose evaluations sum is equal to 'maxi'.
    # For each solution
    for i in range(len(values[0])) :
        sum = 0   # To keep the sum of attendees'evaluations.
        # For each attendee
        for j in range(len(values)) :
            if values[j][i] == 'impossible' :
                # Next place.
                break
            elif values[j][i] == 'negative' :
                sum = sum - 1
            elif values[j][i] == 'positive' :
                sum = sum + 1
        else :
            if (maxi == None) or (maxi < sum) :
                maxi = sum
                maxon = i

    if maxi != None :
        nego_node.setAttributeNS(PIA_NS,PIA_PREFIX+'status', 'accepted')
        comp_node = doc.createElementNS(PIA_NS, 'compromise')
        outputs.appendChild(comp_node)
        set = _discretize(vevent_node)
        comp_node.appendChild(set[maxon].cloneNode(DEEP_COPY))
        vevent = select_one(comp_node, 'VEVENT')
        summary = select_one(vevent, 'SUMMARY/text()')
        summary.data = re.sub('XXX ', '', summary.data)
        if not summary.data:
            summary.data = u'No subject'
        _release_list(set)
        # Delete DTINTERVAL*
        startnode = select_one(vevent,'extension[@x-name="X-NARVAL-DTINTERSTART"]')
        starttime = select_one(startnode, 'text()').data
        vevent.removeChild(startnode)
        release_domtree(startnode)
        duranode = select_one(vevent, 'extension[@x-name="X-NARVAL-DURATION"]')
        vevent.removeChild(duranode)
        release_domtree(duranode)
        endnode = select_one(vevent,'extension[@x-name="X-NARVAL-DTINTEREND"]')
        endtime = select_one(endnode, 'text()').data
        vevent.removeChild(endnode)
        release_domtree(endnode)
        # Create DTSTART and DTEND nodes
        dtstart = doc.createElement('DTSTART')
        vevent.appendChild(dtstart)
        text = doc.createTextNode(starttime)
        dtstart.appendChild(text)
        dtend = doc.createElement('DTEND')
        vevent.appendChild(dtend)
        text = doc.createTextNode(endtime)
        dtend.appendChild(text)
        print 'Compromise found:', starttime, 'to', endtime
    else :
        print 'No compromise found'
        nego_node.setAttributeNS(PIA_NS,PIA_PREFIX+'status', 'failed')
    return outputs

MOD_XML = MOD_XML + \
          """<action name='interpret_evaluation' func='interpret_evaluation_f'>
          <description lang="fr">Dcider si la proposition value prcdemment par les invits est satisfaisante ou non.</description>
          <description lang="en">Decides if previous proposition evaluated by attendees solve the negociation or not.</description>
          <input id='nego'><match>negociation[@subject='event scheduling' and @status='to evaluate']</match></input>
          <output outdates='nego'><match>negociation</match></output>
          <output optional='yes'><match>compromise</match></output>
          </action>
          """


def new_event_proposal_f(args) :
    '''Function creating a new proposal to be sent by the initiator when the last one
    did not please to the agents.
    '''
    doc = args.ownerDocument
    outputs = doc.createElementNS(PIA_NS,'outputs')

    prop_node = select_one(args,'negociation/cycle[last()]/proposal')

    startnode = select_one(prop_node, 'VEVENT/extension[@x-name="X-NARVAL-DTINTERSTART"]/text()')
    endnode = select_one(prop_node, 'VEVENT/extension[@x-name="X-NARVAL-DTINTEREND"]/text()')
    oldstart = startnode.data
    oldend = endnode.data
    start = oldend
    end = _time_to_vcal(2 * _vcal_to_time(oldend) - _vcal_to_time(oldstart))
    startnode.data = start
    endnode.data = end

    tags = select_many(prop_node, 'tags')
    if tags :
        prop_node.removeChild(tags[0])
        release_domtree(tags[0])

    print 'New proposal:', start, end

    outputs.appendChild(prop_node)
    return outputs

MOD_XML = MOD_XML + \
          """<action name='new_event_proposal' func='new_event_proposal_f'>
          <description lang="fr">Crer une nouvelle proposition d'horaire aprs l'chec de la prcdente.</description>
          <description lang="en">Makes a new proposal for scheduling event after the previous one failed.</description>
          <input use='yes'><match>negociation[@subject='event scheduling' and @status='failed']</match></input>
          <output><match>proposal</match></output>
          </action>
          """

MOD_XML = MOD_XML + "</module>"

if __name__ == "__main__" :
    from xml.dom.ext.reader import Sax2
    print MOD_XML
    doc=Sax2.FromXml(MOD_XML)
    print doc.documentElement

